# Etcetera #

![etcetera-logo.png](etcetera-logo.png)

**Config file management with a touch of wisdom.**

## About ##

Etcetera is a simple command line tool that lets you keep track of changes to
config files of a linux system.
It is loosely inspired from the basic functionality of git (commit and revert).

## Why Etcetera ##

When I configure or troubleshoot a system I make sometimes a lot of changes to
config files, some good, some that need to be reverted because they do not have
the intended effect. In the heat of the action, it is easy to lose track of
what change has been made to what file.

I started Etcetera to address that issue. With it I keep a copy of the untouched
original file and I preserve changes that I make associated with a note that
helps me remember what I have done. And just by looking at what file Etcetera
manages, I can tell exactly what packages use configuration that deviates from
installation defaults.

## Features ##

* Strictly define locations where files can be managed by Etcetera
* Select which files to manage
* preserve changes made to a managed file with associated note
* Restore changes by selecting a previous preserved version
* Restore managed file back to the original location
* Colorized output (off by default)

## Download and install ##

### Archlinux ###

The `etcetera` package is available in the [AUR](https://aur.archlinux.org/packages/etcetera).

### Void Linux ###

    xbps-install -S etcetera

### Manual installation ###

Clone the project or download and unpack the zip file from github.

First `cd` into the directory where you unpacked the zip

Make sure `./install-etcetera.sh` is executable:

    $ sudo chmod u+x ./install-etcetera.sh

Then run:

    $ sudo ./install-etcetera.sh

To uninstall later, run:

    $ sudo /usr/share/install-etcetera.sh --remove

For information, the install script copies project files to the following
locations on your system:

| File                | Location                      | Mode |
|---------------------|-------------------------------|------|
| `etcetera`          | `/usr/bin`                    | 755  |
| `parser.py`         | `/usr/lib/etcetera`           | 755  |
| `commands.py`       | `/usr/lib/etcetera`           | 640  |
| `toolbox.py`        | `/usr/lib/etcetera`           | 640  |
| `term_colors.py`    | `/usr/lib/etcetera`           | 640  |
| `README.md`         | `/usr/share/etcetera`         | 640  |
| `etcetera-logo.png` | `/usr/share/etcetera`         | 640  |
| `etcetera.conf`     | `/etc/etcetera.conf`          | 640  |

It makes sure they are all owned by `root:root` and have the correct mode set.

Etcetera has an optional dependency on the
[cysystemd python module](https://github.com/mosquito/cysystemd) to display
status and error messages in the system journal (journald). If you want this
functionality, you can install the module with pip:

    $ sudo pip install cysystemd

> Note : The *original location* (where the files that you want to manage
  are located) and the *managed location* (where the files generated by
  Etcetera are located) must be on the same disk partition. That is
  because Etcetera uses hard links to the managed files. By default the
  `ORIGINAL_LOCATION` is set to `/etc` and `MANAGED_LOCATION` is set to
  `/var/lib/etcetera`. These locations are both configurable in 
  '/etc/etcetera.conf'.

## Documentation ##

### Using Commands ###

Let's assume that we want to manage and track the configuration of the
network time protocol `/etc/ntp.conf` (I chose it only because the file name
is short...).

#### Initialization ####

The management of a given file needs to be initialized first with the command:

    $ etcetera --manage /etc/ntp.conf
    SUCCESS: Copy of /etc/ntp.conf set up in managed location

This command will create a hard link to the file under the directory managed 
by Etcetera (by default `/var/lib/etcetera`, defined under `MANAGED_LOCATION`
in the configuration file `/etc/etcetera.conf`).

All Etcetera commands for which a file name must be provided require a full
path to the file (i.e. starting from the root like `/etc/ntp.conf` and not
just `ntp.conf`).

If the provided path is a directory instead of a file, all files in this 
directory and its sub-directories will be managed.

> NOTE: There are abbreviations for all commands. `--manage` can be replaced by `-m`.
  Please execute `etcetera --help` for more details.

> NOTE: Etcetera will only accept to manage files that are located under an
  allowed location set in `/etc/etcetera.conf` under `ORIGINAL_LOCATIONS`. By
  default, only `/etc` is allowed but more locations can be added (space-separated 
  list).

To verify at any point if the file is managed by Etcetera, run:

    $ etcetera --status /etc/ntp.conf
    File was preserved on these dates:
        1 | Sat Jan 20 14:43:03 2018 | root root | -rw-r--r-- | (original file)
    File is the same as last preserved version

This states that the original version of file is now preserved and that the
current version of the file is identical to the preserved version. User, group
and mode is displayed for each version

#### Modifications of the content of file ####

Once the initialization is successful, all operations that are typically
performed on a config file (editing, patching, merging, ...) can continue to
be done. for example, if you are moving to Europe you can edit the file
to adjust the first ntp pool:

    $ sudo nano /etc/ntp.conf

add `server europe.pool.ntp.org` as first server entry and save (CTRL+X / Y).

If the status command is executed again, the output is slightly different:

    $ etcetera --status /etc/ntp.conf
    File was preserved on these dates:
        1 | Sat Jan 20 14:43:03 2018 | root root | -rw-r--r-- | (original file)
    Some changes to the file are not preserved

The last sentence informs that there are now some differences between the
current version of the file and the last preserved version.

#### Preserving the current version of the file ####

To preserve this version of the file to be able to restore it at a later point
in time:

    $ etcetera --preserve /etc/ntp.conf "Config for Europe"
    SUCCESS: version preserved

The second paramater is a one-line description of what the changes are about.
This parameter is mandatory by default but this requirement can be disabled
by setting `PRESERVE_NOTE_REQUIRED = false` in `/etc/etcetera.conf`.

The status command will show what changed:

    $ etcetera --status /etc/ntp.conf
    File was preserved on these dates:
        2 | Thu May 17 22:41:04 2018 | root root | -rw-r--r-- | Config for Europe
        1 | Sat Jan 20 14:43:03 2018 | root root | -rw-r--r-- | (original file)
    File is the same as last preserved version

If the total number of saved files exceeds the maximum permitted
(`PRESERVE_MAX_SAVES = 5` by default in `/etc/etcetera.conf`), the oldest
preserved version is deleted. The original file is preserved though.

#### Displaying differences between current and preserved versions of the file ####

To see what differences exist inside the file between the current version and
one of the preserved versions use the `--changes` command.

    $ etcetera --changes /etc/ntp.conf
    File was preserved on these dates:
        1 | Sat Jan 20 14:43:03 2018 | root root | -rw-r--r-- | (original file)
    Some changes to the file are not preserved

    Select file version to display changes from (1-1, 0 to abort): 1

The diff command that is executed can be configured in `etcetera.conf` under
`DIFF_COMMAND` in the `[BEHAVIOR]` section. By default it is
`diff -uN {} {}` where braces are placeholders for file paths (preserved file
first, managed file second). This will display a summary of changes. Using
`diff -yN {} {}` instead will display the full files side-by-side.

#### Restoring a preserved version of the file ####

To replace the current version of the file by one that was saved earlier,
in this case the original file, execute:

    $ etcetera --restore /etc/ntp.conf
    File was preserved on these dates:
        2 | Thu May 17 22:41:04 2018 | root root | -rw-r--r-- | Config for Europe
        1 | Sat Jan 20 14:43:03 2018 | root root | -rw-r--r-- | (original file)
    Select file version to restore to (1-2, 0 to abort): 1
    SUCCESS: selected version restored

The status command will show:

    $ etcetera --status /etc/ntp.conf
    File was preserved on these dates:
        2 | Thu May 17 22:41:04 2018 | root root | -rw-r--r-- | Config for Europe
        1 | Sat Jan 20 14:43:03 2018 | root root | -rw-r--r-- | (original file)
    Some changes to the file are not preserved

With the last sentence confirming that the current version (now identical to the
original file) is different from the content of the last preserved version
(the configuration with the European server).

#### Edit the note associated with a preserved version ####

To change the comment use the following command:

    $ etcetera --edit /etc/ntp.conf
    File was preserved on these dates:
        2 | Thu May 17 22:41:04 2018 | root root | -rw-r--r-- | Config for Europe
        1 | Sat Jan 20 14:43:03 2018 | root root | -rw-r--r-- | (original file)
    Some changes to the file are not preserved

    Select version to edit comment for (1-2, 0 to abort): 2
    Press [CTRL]+c to abort
    Old comment : Config for Europe
    New comment : Use EU server 
    SUCCESS: comment modified

### Delete a preserved version ###

To delete a specific preserved version of the managed file use:

    $ etcetera --delete /etc/ntp.conf
    File was preserved on these dates:
        2 | Thu May 17 22:41:04 2018 | root root | -rw-r--r-- | Config for Europe
        1 | Sat Jan 20 14:43:03 2018 | root root | -rw-r--r-- | (original file)
    Some changes to the file are not preserved

    Select version to delete (1-2, 0 to abort): 2
    SUCCESS: version deleted

#### List files managed by Etcetera ####

    $ etcetera --list
    Files managed by etcetera:
     /etc/ntp.conf
    Number of files managed: 1

#### Stop managing file with Etcetera ####

To stop managing a file and delete all the preserved versions, run:

    $ etcetera --unmanage /etc/ntp.conf
    WARNING: This will delete all preserved copies of /etc/ntp.conf
    WARNING: Are you really sure? (y/N): y
    SUCCESS:Managed content related to /etc/ntp.conf deleted

The preserved files created by Etcetera are deleted and if `UNMANAGE_RESTORE_ORIG = true` 
is set in `/etc/etcetera.conf` (it is `false` by default) a copy of the original
version of the file is also placed in the original location with the `.orig`
extension. In this case, `/etc/ntp.conf.orig` will appear alongside
`/etc/ntp.conf`. If there is no change between the managed file and the original
file the .orig file will be omitted.

If the provided path is a directory instead of an individual managed file, Etcetera
will stop managing all files and delete all the related preserved files in this directory
and its sub-directories.

#### Get some information about the configuration and state of Etcetera ####

Use:

    $ etcetera --info
    Location of config file:
     /etc/etcetera.conf
    Locations where files can be monitored:
     /etc
    Location of managed files:
     /var/lib/etcetera
    Files managed by etcetera:
    Number of files managed: 0
    Number of files with changes that are not preserved: 0
    Max. number of backups preserved for each file: 5
    Original files are preserved: True

> NOTE: The output of this command is preliminary and will change in the future.

### Terminology ###

The folder in which it is allowed to select files for management with Etcetera
is called the **original location** (`/etc` by default). More than one such
folder can be specified in the configuration file of Etcetera.

The file that you have chosen to manage, for example `/etc/ntp.conf` is called
the **original file**.

| Path            | Name              |
|-----------------|-------------------|
| `/etc/`         | Original location |
| `/etc/ntp.conf` | Original file     |

#### After call to --manage ####

Once you have called the *manage* command, The managed file is moved to a
directory managed by Etcetera called the **managed location**
(`/var/lib/etcetera` by default). The managed file is preserved and a hard link
to it is created, it is called the **Linked File**.

Every time a successful call to -- manage or --preserve is made, a copy of the
managed file is made that is used to detect whether a change has been made to
the managed file that need to be preserved. This file is called the **shadow
file** and in our example it is `/var/lib/etcetera/etc/ntp.conf.COMMIT`.

| Path                                    | Name             |
|-----------------------------------------|------------------|
| `/etc/ntp.conf`                         | Managed File     |
| `/var/lib/etcetera`                     | Managed Location |
| `/var/lib/etcetera/etc/ntp.conf`        | Linked File      |
| `/var/lib/etcetera/etc/ntp.conf.ORIG`   | Original File    |
| `/var/lib/etcetera/etc/ntp.conf.COMMIT` | Shadow File      |

#### After call(s) to --preserve ####

Each call to --preserve creates a pair of files with an extension including a
timestamp. One is called the **commit file**
(`/var/lib/etcetera/etc/ntp.conf.COMMIT_YYYY-MM-DD_HH-MM-SS`) and the other the
 **comment file** (`/var/lib/etcetera/etc/ntp.conf.COMMENT_YYYY-MM-DD_HH-MM-SS`)

| Path                                                         | Name         |
|--------------------------------------------------------------|--------------|
| `/var/lib/etcetera/etc/ntp.conf.COMMIT_YYYY-MM-DD_HH-MM-SS`  | commit file  |
| `/var/lib/etcetera/etc/ntp.conf.COMMENT_YYYY-MM-DD_HH-MM-SS` | comment file |

All the files in the managed location that share the same base name as the
managed file are called the **associated files** (`*.ORIG`, `*.COMMIT`,
`*.COMMIT_YYYY-MM-DD_HH-MM-SS` and `*.COMMENT_YYYY-MM-DD_HH-MM-SS`).

## Roadmap ##

[roadmap.md](roadmap.md)

## Contributors ##

Jean-Christophe Francois <jc.francois@gmail.com>

## License ##

### GNU General Public License version 3 ###

    Copyright (C) 2019  Jean-Christophe Francois

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

Full terms and conditions are available in the bundled [LICENSE](LICENSE) file.
